from Cryptodome.Cipher import AES
from ledgerblue.ecWrapper import PrivateKey
from ecpy.curves import Curve
import hashlib
import struct


def decrypt_certificate(encrypted_certificate, key):
    tag = encrypted_certificate[:16]
    ciphertext = encrypted_certificate[16:]
    aes_siv = AES.new(key, mode=AES.MODE_SIV)
    certificate = aes_siv.decrypt_and_verify(ciphertext, tag)
    return certificate


def extract_from_certificate(certificate):
    offset = 1
    certificate_header = certificate[offset: offset + certificate[offset - 1]]
    offset += certificate[offset - 1] + 1
    certificate_public_key = certificate[offset: offset + certificate[offset - 1]]
    offset += certificate[offset - 1] + 1
    certificate_signature_array = certificate[offset: offset + certificate[offset - 1]]
    return certificate_header, certificate_public_key, certificate_signature_array


def scp_derive_key(ecdh_secret, keyindex, scpv3=False):
    if scpv3:
        mac_block = b'\x01' * 16
        cipher = AES.new(ecdh_secret, AES.MODE_ECB)
        mac_key = cipher.encrypt(mac_block)
        enc_block = b'\x02' * 16
        cipher = AES.new(ecdh_secret, AES.MODE_ECB)
        enc_key = cipher.encrypt(enc_block)
        return mac_key + enc_key
    retry = 0
    # di = sha256(i || retrycounter || ecdh secret)
    while True:
        sha256 = hashlib.new('sha256')
        sha256.update(struct.pack(">IB", keyindex, retry))
        sha256.update(ecdh_secret)

        # compare di with order
        CURVE_SECP256K1 = Curve.get_curve('secp256k1')
        if int.from_bytes(sha256.digest(), 'big') < CURVE_SECP256K1.order:
            break
        # regenerate a new di satisfying order upper bound
        retry += 1

    # Pi = di*G
    privkey = PrivateKey(bytes(sha256.digest()))
    pubkey = bytearray(privkey.pubkey.serialize(compressed=False))
    # ki = sha256(Pi)
    sha256 = hashlib.new('sha256')
    sha256.update(pubkey)
    return sha256.digest()




"""
*******************************************************************************
*   Ledger Blue
*   (c) 2023 Ledger
*
*  Licensed under the Apache License, Version 2.0 (the "License");
*  you may not use this file except in compliance with the License.
*  You may obtain a copy of the License at
*
*      http://www.apache.org/licenses/LICENSE-2.0
*
*  Unless required by applicable law or agreed to in writing, software
*  distributed under the License is distributed on an "AS IS" BASIS,
*  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*  See the License for the specific language governing permissions and
*  limitations under the License.
********************************************************************************
"""

import argparse
import os
from contextlib import contextmanager
from elftools.elf.elffile import ELFFile

__ELF_METADATA_SECTIONS = [
    "target",
    "target_name",
    "target_id",
    "app_name",
    "app_version",
    "api_level",
    "sdk_version",
    "sdk_name",
    "sdk_hash",
]

@contextmanager
def _get_elf_file(filename):
    if os.path.exists(filename):
        with open(filename, 'rb') as fp:
            yield ELFFile(fp)
    else:
        raise FileNotFoundError(f"File {filename} does not exist.")


def _get_elf_section_value(elf, section_name):
    section = elf.get_section_by_name(f"ledger.{section_name}")
    section_value = ""
    if section:
        section_value = section.data().decode("utf-8").strip()
    return section_value


def get_elf_section_value(filename, section_name):
    with _get_elf_file(filename) as elf:
        return _get_elf_section_value(elf, section_name)


def get_target_id_from_elf(filename):
    return get_elf_section_value(filename, "target_id")


def get_argparser():
    parser = argparse.ArgumentParser(
        description="""Read the metadata of a Ledger device's ELF binary file.""")
    parser.add_argument(
        "--fileName", help="The name of the ELF binary file to read", required=True)
    parser.add_argument(
        "--section", help=f"The name of the metadata section to be read. If no value is provided, all sections are read.", choices=__ELF_METADATA_SECTIONS + ["all"], default="all")
    return parser


if __name__ == '__main__':

    args = get_argparser().parse_args()

    with _get_elf_file(args.fileName) as elf:
        if(args.section == "all"):
            for section_name in __ELF_METADATA_SECTIONS:
                section_value = _get_elf_section_value(elf, section_name)
                print(f"{section_name} : {section_value}")
        else:
            print(_get_elf_section_value(elf, args.section))      



from ecpy.curves import Curve, Point

class PedersenVSS:

  def __init__(self, curve: Curve):
    self.curve = curve
    self.P = Point(curve.generator.x, curve.generator.y, curve)
    self.domain_len = curve.generator.x.bit_length() // 8

  def pedersen_commit(self, Q: Point, a: int, b: int) -> Point:
    point1 = self.curve.mul_point(a, self.P)
    point2 = self.curve.mul_point(b, Q)

    return self.curve.add_point(point1, point2)

  def pedersen_share_commit(self, Q: Point, share: bytes) -> Point:
    a = int.from_bytes(share[:self.domain_len], 'big')
    b = int.from_bytes(share[self.domain_len:], 'big')
    return self.pedersen_commit(Q, a, b)


  def pedersen_verify_commit(self, Q: Point, share: bytes, index: int, commits: list):
    s_point = self.pedersen_share_commit(Q, share)

    r_point = commits[0]

    for i in range(1, len(commits)):
      r = self.curve.mul_point(index**i, commits[i])
      r_point = self.curve.add_point(r_point, r)

    return s_point == r_point, s_point





"""
*******************************************************************************
*   Ledger Blue
*   (c) 2016 Ledger
*
*  Licensed under the Apache License, Version 2.0 (the "License");
*  you may not use this file except in compliance with the License.
*  You may obtain a copy of the License at
*
*      http://www.apache.org/licenses/LICENSE-2.0
*
*  Unless required by applicable law or agreed to in writing, software
*  distributed under the License is distributed on an "AS IS" BASIS,
*  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*  See the License for the specific language governing permissions and
*  limitations under the License.
********************************************************************************
"""

import argparse

def get_argparser():
	parser = argparse.ArgumentParser(description="Set a BIP 39 passphrase on the device.")
	parser.add_argument("--persistent", help="""Persist passphrase as secondary PIN (otherwise, it's set as a temporary
passphrase)""", action='store_true')
	return parser

def auto_int(x):
	return int(x, 0)

if __name__ == '__main__':
	from .comm import getDongle
	import getpass
	import unicodedata
	import sys

	args = get_argparser().parse_args()

	dongle = getDongle(False)

	passphrase = getpass.getpass("Enter BIP39 passphrase : ")
	if isinstance(passphrase, bytes):
		passphrase = passphrase.decode(sys.stdin.encoding)
	if len(passphrase) != 0:
		if args.persistent:
			p1 = 0x02
		else:
			p1 = 0x01
		passphrase = unicodedata.normalize('NFKD', passphrase)
		apdu = bytearray([0xE0, 0xD0, p1, 0x00, len(passphrase)]) + bytearray(passphrase, 'utf8')
		dongle.exchange(apdu, timeout=300)

	dongle.close()



"""
*******************************************************************************
*   Ledger Blue
*   (c) 2016 Ledger
*
*  Licensed under the Apache License, Version 2.0 (the "License");
*  you may not use this file except in compliance with the License.
*  You may obtain a copy of the License at
*
*      http://www.apache.org/licenses/LICENSE-2.0
*
*  Unless required by applicable law or agreed to in writing, software
*  distributed under the License is distributed on an "AS IS" BASIS,
*  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*  See the License for the specific language governing permissions and
*  limitations under the License.
********************************************************************************
"""

from ledgerblue.comm import Dongle
from ledgerblue.commException import CommException
from ledgerblue.ledgerWrapper import wrapCommandAPDU, unwrapResponseAPDU
from binascii import hexlify
from bluepy import btle

SERVICE_UUID = "D973F2E0-B19E-11E2-9E96-0800200C9A66"
WRITE_CHARACTERISTICS_UUID = "D973F2E2-B19E-11E2-9E96-0800200C9A66"
NOTIFY_CHARACTERISTICS_UUID = "D973F2E1-B19E-11E2-9E96-0800200C9A66"
CLIENT_CHARACTERISTICS_CONFIG_DESCRIPTOR_UUID = "00002902-0000-1000-8000-00805f9b34fb"
ENABLE_NOTIFICATION = "0100".decode('hex')
DEFAULT_BLE_CHUNK = 20
BLE_NOTIFICATION_TIMEOUT = 5.0

class BLEDongleDelegate(btle.DefaultDelegate):
	def __init__(self, dongle):
		btle.DefaultDelegate.__init__(self)
		self.dongle = dongle

	def handleNotification(self, cHandle, data):		
		self.dongle.result += bytearray(data)

class BLEDongle(Dongle):

	# Must be called with the Client Characteristics Configuration descriptor handle as bluepy fails to retrieve it
	# From gatttools
	# [device mac][LE]> char-desc
  # handle: 0x000f, uuid: 00002902-0000-1000-8000-00805f9b34fb
	def __init__(self, bleAddress, configDescriptor, debug=False):
		self.device = btle.Peripheral(bleAddress)		
		self.device.setDelegate(BLEDongleDelegate(self))
		self.service = self.device.getServiceByUUID(SERVICE_UUID)
		self.writeCharacteristic = self.service.getCharacteristics(forUUID=WRITE_CHARACTERISTICS_UUID)[0]
		self.device.writeCharacteristic(configDescriptor, ENABLE_NOTIFICATION, withResponse=True)
		self.debug = debug
		self.opened = True

	def exchange(self, apdu, timeout=20000):
		if self.debug:
			print("=> %s" % hexlify(apdu))
		apdu = wrapCommandAPDU(0, apdu, DEFAULT_BLE_CHUNK, True)		
		offset = 0
		while offset < len(apdu):
			data = apdu[offset:offset + DEFAULT_BLE_CHUNK]
			self.writeCharacteristic.write(data, withResponse=True)
			offset += DEFAULT_BLE_CHUNK
			self.result = ""
		while True:
			if not self.device.waitForNotifications(BLE_NOTIFICATION_TIMEOUT):
				raise CommException("Timeout")
			response = unwrapResponseAPDU(0, self.result, DEFAULT_BLE_CHUNK, True)
			if response is not None:
				result = response
				dataStart = 0
				swOffset = len(response) - 2
				dataLength = len(response) - 2
				break
		sw = (result[swOffset] << 8) + result[swOffset + 1]
		response = result[dataStart : dataLength + dataStart]
		if self.debug:
			print("<= %s%.2x" % (hexlify(response), sw))
		if sw != 0x9000:
			raise CommException("Invalid status %04x" % sw, sw)
		return response

	def close(self):
		if self.opened:
			try:
				self.device.disconnect()
			except:
				pass
		self.opened = False



from ledgerblue.ecWrapper import PrivateKey
from ledgerblue.comm import getDongle
from ledgerblue.hexLoader import HexLoader
from ledgerblue.recoverMutualAuth import SCPv3
import binascii
import argparse


def get_argparser():
    parser = argparse.ArgumentParser(description="Delete the custom Certificate Authority used to "
                                                 "backup/restore a seed.")
    parser.add_argument("--targetId", help="The device's target ID (default is Nano X)", type=auto_int)
    parser.add_argument("--name", help="The certificate's name", required=True)
    parser.add_argument("--issuerPublicKey", help="The public key of the Issuer (used to verify the certificate of "
                                                  "the device)")
    parser.add_argument("--rootPrivateKey", help="The private key of the Signer used to establish a secure channel "
                                                 "(otherwise a random one will be generated)")
    parser.add_argument("--caPublicKey", help="The Custom CA's public key to be enrolled (hex encoded)", required=True)
    return parser


def auto_int(x):
    return int(x, 0)


if __name__ == '__main__':

    args = get_argparser().parse_args()

    if args.targetId is None:
        args.targetId = 0x33000004
    if args.name is None:
        raise Exception("Missing certificate's name")
    if args.issuerPublicKey is None:
        args.issuerPublicKey = '0490f5c9d15a0134bb019d2afd0bf297149738459706e7ac5be4abc350a1f81805' \
                               '7224fce12ec9a65de18ec34d6e8c24db927835ea1692b14c32e9836a75dad609'
    if args.rootPrivateKey is None:
        privateKey = PrivateKey()
        publicKey = binascii.hexlify(privateKey.pubkey.serialize(compressed=False))
        print("Generated random root public key : %s" % publicKey)
        args.rootPrivateKey = privateKey.serialize()
    if args.caPublicKey is None:
        raise Exception("Missing CA's public key")

    publicKey = bytearray.fromhex(args.caPublicKey)
    dongle = getDongle(True)

    # Execute secure channel protocol (new version)
    secret, devicePublicKey = SCPv3(dongle, bytearray.fromhex(args.issuerPublicKey), args.targetId,
                                    None, bytearray.fromhex(args.rootPrivateKey),
                                    user_mode=True)
    loader = HexLoader(dongle, 0xe0, True, secret, scpv3=True)

    # Delete the Certificate Authority's public key
    loader.recoverDeleteCA(args.name, publicKey)



# -*- coding: utf-8 -*-
#
# BOLOS Python Loader documentation build configuration file.

import os
import sys

from ledgerblue.__version__ import __version__

def setup(app):
    app.add_css_file('theme_overrides.css') # Override wide tables in RTD theme

# General Configuration
# =====================

extensions = []

source_suffix = ['.rst']

master_doc = 'index'

project = u'BOLOS Python Loader'
copyright = u'2017, Ledger Team'
author = u'Ledger Team'

version = __version__
release = __version__

pygments_style = 'sphinx'

# Options for HTML Output
# =======================

html_theme = 'sphinx_rtd_theme'

html_static_path = ['_static']

# sphinxarg
# =========

extensions += ['sphinxarg.ext']



"""
*******************************************************************************
*   Ledger Blue
*   (c) 2016 Ledger
*
*  Licensed under the Apache License, Version 2.0 (the "License");
*  you may not use this file except in compliance with the License.
*  You may obtain a copy of the License at
*
*      http://www.apache.org/licenses/LICENSE-2.0
*
*  Unless required by applicable law or agreed to in writing, software
*  distributed under the License is distributed on an "AS IS" BASIS,
*  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*  See the License for the specific language governing permissions and
*  limitations under the License.
********************************************************************************
"""

import argparse

def get_argparser():
	parser = argparse.ArgumentParser("Update the firmware by using Ledger to open a Secure Channel.")
	parser.add_argument("--url", help="Websocket URL", default="wss://scriptrunner.api.live.ledger.com/update/install")
	parser.add_argument("--bypass-ssl-check", help="Keep going even if remote certificate verification fails", action='store_true', default=False)
	parser.add_argument("--apdu", help="Display APDU log", action='store_true')
	parser.add_argument("--perso", help="""A reference to the personalization key; this is a reference to the specific
Issuer keypair used by Ledger to sign the device's Issuer Certificate""", default="perso_11")
	parser.add_argument("--firmware", help="A reference to the firmware to load", required=True)
	parser.add_argument("--targetId", help="The device's target ID (default is Ledger Blue)", type=auto_int, default=0x31000002)
	parser.add_argument("--firmwareKey", help="A reference to the firmware key to use", required=True)
	return parser

def auto_int(x):
	return int(x, 0)

def process(dongle, request):
	response = {}
	apdusList = []
	try:
		response['nonce'] = request['nonce']
		if request['query'] == "exchange":
			apdusList.append(binascii.unhexlify(request['data']))
		elif request['query'] == "bulk":
			for apdu in request['data']:
				apdusList.append(binascii.unhexlify(apdu))
		else:
			response['response'] = "unsupported"
	except:
		response['response'] = "parse error"

	if len(apdusList) != 0:
		try:
			for apdu in apdusList:
				response['data'] = dongle.exchange(apdu).hex()
			response['response'] = "success"
		except:
			response['response'] = "I/O" # or error, and SW in data

	return response

if __name__ == '__main__':
	import urllib.parse as urlparse
	from .comm import getDongle
	from websocket import create_connection
	import json
	import binascii
	import ssl

	args = get_argparser().parse_args()

	dongle = getDongle(args.apdu)

	url = args.url
	queryParameters = {}
	queryParameters['targetId'] = args.targetId
	queryParameters['firmware'] = args.firmware
	queryParameters['firmwareKey'] = args.firmwareKey
	queryParameters['perso'] = args.perso
	queryString = urlparse.urlencode(queryParameters)
	if args.bypass_ssl_check:
		# SEE: https://docs.python.org/3/library/ssl.html#ssl.CERT_NONE
		# According to the documentation:
		# > With client-side sockets, just about any cert is accepted. Validation errors, such
		# > as untrusted or expired cert, are ignored and do not abort the TLS/SSL handshake.
		sslopt = { "cert_reqs": ssl.CERT_NONE }
	else:
		sslopt = {}
	ws = create_connection(args.url + '?' + queryString, sslopt=sslopt)
	while True:
		result = json.loads(ws.recv())
		if result['query'] == 'success':
			break
		if result['query'] == 'error':
			raise Exception(result['data'] + " on " + result['uuid'] + "/" + result['session'])
		response = process(dongle, result)
		ws.send(json.dumps(response))
	ws.close()

	print("Script executed successfully")

	dongle.close()
